/* 
   cvsd - chroot wrapper to run `cvs pserver' more securely.

   cvsd was originally written by Chris Black <cblack@mokey.com>,
   http://cblack.mokey.com/. That was until release 0.6

   cvsd versions up till 0.8b3 were maintained by Philippe Kehl
   <phkehl@gmx.net>, http://guv.ethz.ch/~flip/cvsd/,
   http://www.oinkzwurgl.org/software/cvsd/.

   after that Arthur de Jong <arthur@tiefighter.et.tudelft.nl>
   took up the work and did a complete rewrite,
   http://tiefighter.et.tudelft.nl/~arthur/cvsd.

   Copyright (C) 1999 Chris Black.
   Copyright (C) 2000 Philippe Kehl.
   Copyright (C) 2001 Arthur de Jong.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
*/


#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>
#include <getopt.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/stat.h>


#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>
#include <getopt.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "config.h"


/* the name by which this program was run. */
char *program_name;


/* the command that will be run */
static char *cvs_cmd="/bin/cvs";


/* the arguments that will be passed */
static char **cvs_args;


/* the environment cvs will run in */
static char *cvs_env[]={
  "HOME=/",
  "TMPDIR=/tmp",
  "CVSUMASK=027", /* i think this one's important. how strictly does cvs follow this? */
  NULL
};


/* the umask value cvs will run as */
static mode_t cvs_umask=0027;


/* the configuration file default from Makefile */
char *configfile=DEFAULT_CONFIGFILE;


/* the user id cvs should be run as */
#define NOUID ((gid_t)-1)
uid_t uid=NOUID;


/* the group id cvs should be run as */
#define NOGID ((gid_t)-1)
gid_t gid=NOGID;


/* cvsd root directory */
char *rootjail=NULL;


/* temporary linked list of repos diretories */
struct repos_type
{
  char *repos;
  struct repos_type *next;
} *cvsrepos=NULL;
int cvsrepos_size=0;


/* the port to be used for listening */
int port=2401;


/* the maximum number of connections to handle */
int maxconnections=0; /* default=0=unlimited */


/* the number of children spawned */
static int numchildren=0;


/* the number of seconds to sleep when no more
   connections can be listened for */
#define SLEEPSECS 5


/* maximum linelength in the configuration file */
#define MAX_LINE_LENGTH 1024


/* display version information */
static void display_version(FILE *fp)
{
  fprintf(fp,"%s %s\n",PACKAGE,VERSION);
  fprintf(fp,"Written by Chris Black, Philippe Kehl and Arthur de Jong.\n\n");
  fprintf(fp,"Copyright (C) 2000,2001 Chris Black, Philippe Kehl and Arthur de Jong.\n"
             "This is free software; see the source for copying conditions.  There is NO\n"
             "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");
}


/* display usage information to stdout and exit(status) */
void display_usage(FILE *fp)
{
  fprintf(fp,"Usage: %s [OPTION]...\n", program_name);
  fprintf(fp,"chroot wrapper to run `cvs pserver' more securely.\n");
  fprintf(fp,"  -f, --config=FILE  use FILE as configfile (default %s)\n",DEFAULT_CONFIGFILE);
  fprintf(fp,"      --help         display this help and exit\n");
  fprintf(fp,"      --version      output version information and exit\n");
  fprintf(fp,"\n"
             "Report bugs to <arthur@tiefighter.et.tudelft.nl>.\n");
}


/* the definition of options for getopt(). see getopt(2) */
static struct option const cvsd_options[] =
{
  { "config",      required_argument, NULL, 'f' },
  { "help",        no_argument,       NULL, 'h' },
  { "version",     no_argument,       NULL, 'V' },
  { NULL, 0, NULL, 0 }
};
#define CVSD_OPTIONSTRING "f:hV"


/* parse command line options and set global vars */
void parse_cmdline(int argc, char *argv[])
{
  int optc;
  while ((optc=getopt_long(argc,argv,CVSD_OPTIONSTRING,cvsd_options,NULL))!=-1)
  {
    switch (optc)
    {
    case 'f': /* -f, --config=FILE  use FILE as configfile */
      configfile=(char *)malloc((strlen(optarg)+1)*sizeof(char));
      strcpy(configfile,optarg);
      break;
    case 'h': /*     --help         display this help and exit */
      display_usage(stdout);
      exit(0);
    case 'V': /*     --version      output version information and exit */
      display_version(stdout);
      exit(0);
    case ':': /* missing required parameter */
    case '?': /* unknown option character or extranous parameter */
    default:
      fprintf(stderr,"Try `%s --help' for more information.\n",
              program_name);
      exit(1);
    }
  }
  /* check for remaining arguments */
  if (optind<argc)
  {
    fprintf(stderr,"%s: unrecognized option `%s'\n",
            program_name,argv[optind]);
    fprintf(stderr,"Try `%s --help' for more information.\n",
            program_name);
    exit(1);
  }
}


/* read the configuration file */
void read_configfile(char *filename)
{
  FILE *fp;
  int lnr=0;
  char line[MAX_LINE_LENGTH];
  char *tmp;
  int i;
  struct repos_type *newrepos;

  /* open configuration file */
  if ((fp=fopen(filename,"r"))==NULL)
  {
    syslog(LOG_ERR,"cannot open config file (%s): %s",configfile,strerror(errno));
    exit(1);
  }

#ifdef DEBUG
  syslog(LOG_DEBUG,"reading config file (%s)",configfile);
#endif /* DEBUG */

  /* read all lines in the file */
  while (fgets(line,MAX_LINE_LENGTH,fp)!=NULL)
  {
    lnr++;

    /* strip newline */
    i=strlen(line);
    if ((i>0)&&(line[i-1]=='\n'))
    {
      line[i-1]='\0';
    }
    else
    {
      syslog(LOG_ERR,"%s:%d: line too long or last line missing newline",configfile,lnr);
      exit(1);
    }

    /* strip out comments */
    tmp=strchr(line,'#');
    if (tmp!=NULL) *tmp='\0';

    /* convert tabs to spaces */
    tmp=line;
    while (*tmp!='\0')
    {
      if (*tmp=='\t') *tmp=' ';
      tmp++;
    }

    /* remove trailing spaces */
    tmp--;
    while((tmp>=line)&&(*tmp==' '))
    {
      *tmp='\0';
      tmp--;
    }
    
    /* skip beginning spaces */
    tmp=line;
    while (*tmp==' ')
      tmp++;
    
    /* find identifyer */
    if (strncmp(tmp,"RootJail ",9)==0)
    {
      /* skip space */
      for (tmp+=9;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: RootJail: argument missing",configfile,lnr);
        exit(1);
      }
      if (rootjail==NULL)
      {
        rootjail=(char *)malloc((strlen(tmp)+1)*sizeof(char));
        strcpy(rootjail,tmp);
      }
    }
    else if (strncmp(tmp,"Uid ",4)==0)
    {
      /* skip space */
      for (tmp+=4;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: Uid: argument missing",configfile,lnr);
        exit(1);
      }
      /* TODO: support non-numerical uids */
      uid=strtol(tmp,&tmp,0);
      if (*tmp!='\0')
      {
        syslog(LOG_ERR,"%s:%d: Uid: wrong argument",configfile,lnr);
        exit(1);
      }
    }
    else if (strncmp(tmp,"Gid ",4)==0)
    {
      /* skip space */
      for (tmp+=4;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: Gid: argument missing",configfile,lnr);
        exit(1);
      }
      /* TODO: support non-numerical gids */
      gid=strtol(tmp,&tmp,0);
      if (*tmp!='\0')
      {
        syslog(LOG_ERR,"%s:%d: Uid: wrong argument",configfile,lnr);
        exit(1);
      }
    }
    else if (strncmp(tmp,"Port ",5)==0)
    {
      /* skip space */
      for (tmp+=5;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: Port: argument missing",configfile,lnr);
        exit(1);
      }
      port=strtol(tmp,&tmp,0);
      if (*tmp!='\0')
      {
        syslog(LOG_ERR,"%s:%d: Port: wrong argument",configfile,lnr);
        exit(1);
      }
    }
    else if (strncmp(tmp,"MaxConnections ",15)==0)
    {
      /* skip space */
      for (tmp+=15;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: MaxConnections: argument missing",configfile,lnr);
        exit(1);
      }
      maxconnections=strtol(tmp,&tmp,0);
      if (*tmp!='\0')
      {
        syslog(LOG_ERR,"%s:%d: MaxConnections: wrong argument",configfile,lnr);
        exit(1);
      }
    }
    else if (strncmp(tmp,"Repos ",6)==0)
    {
      /* skip space */
      for (tmp+=6;*tmp==' ';tmp++);
      if (*tmp=='\0')
      {
        syslog(LOG_ERR,"%s:%d: Repos: argument missing",configfile,lnr);
        exit(1);
      }
      /* add repository to list */
      newrepos=(struct repos_type *)malloc(sizeof(struct repos_type));
      newrepos->repos=(char *)malloc((strlen(tmp)+1)*sizeof(char));
      strcpy(newrepos->repos,tmp);
      newrepos->next=cvsrepos;
      cvsrepos=newrepos;
      cvsrepos_size++;
    }
    else if (*tmp!='\0') /* line should be empty */
    {
      syslog(LOG_ERR,"%s:%d: unrecogdnized line %c",configfile,lnr,*tmp);
      exit(1);
    }
  }
  fclose(fp);
}


/* check the configuration */
void check_config()
{
#ifdef DEBUG
  struct repos_type *tmp;
#endif /* DEBUG */

  /* only accept absolute paths for the cvsd-root. abort otherwise. */
  if ((rootjail!=NULL)&&(rootjail[0]!='/'))
  {
    syslog(LOG_ERR,"root (%s) must be an absolute path.",rootjail);
    exit(1);
  }

  /* abort if there was not any valid repository */
  if (cvsrepos_size==0)
  {
    syslog(LOG_ERR,"could not find any valid repository.");
    exit(1);
  }
  
  /* TODO?: check for sane uid/gid values */
  /* TODO?: check for sane port values */
  /* TODO?: check for sane maxconnections value */

  /* some debugging information */
#ifdef DEBUG
    syslog(LOG_DEBUG,"config: uid=%d, gid=%d, rootjail=`%s'",uid,gid,rootjail);
    for (tmp=cvsrepos;tmp!=NULL;tmp=tmp->next)
      syslog(LOG_DEBUG,"config: repos=`%s'",tmp->repos);
    syslog(LOG_DEBUG,"config: port=%d, maxconnections=%d",port,maxconnections);
#endif /* DEBUG */
}


/* signal handler for dieing children */
void sigchld_handler(int signum)
{
  int status;
  int cpid;
  /* TODO?: replace syslog with something else? */
  /* loop until all pending child processes have finished 
     (kill all the zombies) */
  do
  {
    cpid=waitpid(-1,&status,WNOHANG);
    if (cpid>0) /* if there was a terminating child */
    {
      /* shouldn't have a race here beacause we don't have threads */
      numchildren--;
      if (WIFSIGNALED(status))
      {
        syslog(LOG_INFO,"cvsd[%d] exited on signal %d with exit-status %d\n",
                        (int)cpid,
                        (int)WTERMSIG(status),
                        (int)WEXITSTATUS(status));
      }
      else
      {
        syslog(LOG_INFO,"cvsd[%d] exited%s with exit-status %d\n",
                        (int)cpid,
                        WIFEXITED(status)?"":" abnormally",
                        (int)WEXITSTATUS(status));
      }
    }
  }
  while (cpid>0);
}  


/* return a socket for doing accept() on */
int getlistensocket(int port)
{
  int sock;
  struct sockaddr_in addr;

  /* create a socket for tcp connections */
  if ((sock=socket(AF_INET,SOCK_STREAM,0))<0)
  {
    syslog(LOG_ERR,"cannot create socket: %s",strerror(errno));
    exit(1);
  }
  
  /* bind the socket to the specified port */
  addr.sin_family=AF_INET;
  addr.sin_port=htons(port);
  addr.sin_addr.s_addr=INADDR_ANY;
  if (bind(sock,(struct sockaddr *)&addr,sizeof(addr))<0)
  {
    syslog(LOG_ERR,"cannot bind to port %d: %s",port,strerror(errno));
    exit(1);
  }

  /* set some socket options */
/* TODO: this is not working currently */
#ifdef TODO_GETTHISWORKING  
  if (setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,(void *)1,sizeof(SO_REUSEADDR))<0)
  {
    syslog(LOG_ERR,"cannot setsockopt(SO_REUSEADDR): %s",strerror(errno));
    exit(1);
  }
  if (setsockopt(sock,SOL_SOCKET,SO_LINGER,(void *)0,sizeof(SO_LINGER))<0)
  {
    syslog(LOG_ERR,"cannot setsockopt(SO_LINGER): %s",strerror(errno));
    exit(1);
  }
#endif /* TODO_GETTHISWORKING */

  /* start listening for connections */
  if (listen(sock,1)<0)
  {
    syslog(LOG_ERR,"cannot listen on socket: %s",strerror(errno));
    exit(1);
  }
  return sock;
}


/* accept a connection on the socket and spawn a child cvs */
void acceptconnection(int sock)
{
  int csock;
  struct sockaddr_in caddr;
  int clen;
  int cpid;
#ifdef DEBUG
  int i;
#endif /* DEBUG */

  /* accept a new connection */
  clen=sizeof(caddr);
  if ((csock=accept(sock,(struct sockaddr *)&caddr,&clen))<0)
  {
    syslog(LOG_ERR,"accept() failed: %s",strerror(errno));
    exit(1);
  }
  if (clen!=sizeof(caddr))
  {
    syslog(LOG_ERR,"accept() parameter error");
    exit(1);
  }
  
  /* we now have a connection from inet_ntoa(caddr.sin_addr)
     port ntohs(caddr.sin_port) */
  syslog(LOG_INFO,"connection from %s:%d",
                  inet_ntoa(caddr.sin_addr),
                  (int)ntohs(caddr.sin_port));
  
  /* fork off a child to handle the connection */
  switch (cpid=fork())
  {
  case 0: /* we are the child */
    /* connect socket to stdin/stdout/stderr */
    if ( (dup2(csock,0)<0) || /* stdin */
         (dup2(csock,1)<0) || /* stdout */
         (dup2(csock,2)<0) )  /* stderr */
    {
      syslog(LOG_ERR,"dup failed: %s",strerror(errno));
      exit(1);
    }
    /* execute cvs */
    syslog(LOG_INFO,"executing `cvs pserver'");
#ifdef DEBUG
    for (i=0;cvs_args[i]!=NULL;i++)
      syslog(LOG_DEBUG,"cvs_args[%i] = %s",i,cvs_args[i]);
#endif /* DEBUG */
    execve(cvs_cmd,cvs_args,cvs_env);

    /* if we are here there has been an error */
    syslog(LOG_ERR,"error executing `cvs pserver': %s",strerror(errno));
    exit(1);
    break;
  case -1: /* we are the parent, but have an error */
    syslog(LOG_ERR,"fork() failed: %s",strerror(errno));
    exit(1);
    break;
  }

  /* we are the parent and have a cpid */
#ifdef DEBUG
  syslog(LOG_DEBUG,"fork() succeded (child pid=%d)",(int)cpid);
#endif

  /* close the socket to prevent confusion */
  close(csock);
}


/* the main program... */
int main(int argc,char *argv[],char *env[])
{
  int i;
  struct repos_type *tmp;
  int sock;

  /* save program name */
  program_name=argv[0];

  /* parse the command line */
  parse_cmdline(argc,argv);
  
  /* open system logger...
     (from here only use syslog) */
  openlog("cvsd",LOG_PID,LOG_DAEMON);
  syslog(LOG_INFO,"version %s starting",VERSION);

  /* read the configuration file */
  read_configfile(configfile);
  
  /* check the configuration */
  check_config();
  
  /* set the umask */
  umask(cvs_umask);

  /* bind port */
  sock=getlistensocket(port);

  /* try to chdir(there). abort on failure. */
  if ((rootjail!=NULL)&&(chdir(rootjail)!=0))
  {
    syslog(LOG_ERR,"cannot chdir(%s): %s",rootjail,strerror(errno));
    exit(1);
  }

  /* now chroot there. abort on failure. */
  if ((rootjail!=NULL)&&(chroot(rootjail)!=0))
  {
    syslog(LOG_ERR,"cannot chroot(%s): %s",rootjail,strerror(errno));
    exit(1);
  }
  chdir("/");

  /* change to cvs gid */
  if ((gid!=NOGID)&&(setgid(gid))!=0)
  {
    syslog(LOG_ERR,"cannot setgid(%d): %s",(int)gid,strerror(errno));
    exit(1);
  }

  /* change to cvs uid */
  if ((uid!=NOUID)&&(setuid(uid)!=0))
  {
    syslog(LOG_ERR,"cannot setuid(%d): %s",(int)uid,strerror(errno));
    exit(1);
  }

  /* build arguments array for `cvs pserver' */
#ifdef DEBUG
  cvs_args=(char **)malloc((2+cvsrepos_size+1+1)*sizeof(char *));
#else /* not DEBUG */
  cvs_args=(char **)malloc((3+cvsrepos_size+1+1)*sizeof(char *));
#endif /* not DEBUG */
  i=0;
  cvs_args[i++]="cvs";

#ifdef DEBUG
  cvs_args[i++]="-t";  /* verbose */
#endif /* DEBUG */

  cvs_args[i++]="-f"; /* do not read ~/.cvsrc */
  for (tmp=cvsrepos;tmp!=NULL;tmp=tmp->next)
  {
    cvs_args[i]=(char *)malloc((13+strlen(tmp->repos)+1)*sizeof(char));
    strcpy(cvs_args[i],"--allow-root=");
    strcat(cvs_args[i],tmp->repos);
    i++;
  }
  cvs_args[i++]="pserver";
  cvs_args[i++]=NULL;

  /* install signalhandler for terminating children */
  if(signal(SIGCHLD,sigchld_handler)==SIG_ERR)
  {
    syslog(LOG_ERR,"error installing signal handler.");
    exit(1);
  }

  syslog(LOG_INFO,"listening for connections");

  /* start waiting for incoming connections */
  while (1)
  {
    /* if we are ready to handle a connection do so */
    if ((maxconnections>0)&&(numchildren<maxconnections))
    {
      /* wait for a new connection */
      acceptconnection(sock);
      /* we have a possible race here? */
      numchildren++;
    }
    else /* not ready for connection */
    {
      /* sleep onblocks on a signal */
      sleep(SLEEPSECS);
    }
  }

  return 1; /* keep compiler happy */
}
